// *****************************************************************************
//
// Copyright. 2007. Symvionics, Inc. All rights reserved.
// 
// Symvionics Inc. ("Symvionics") grants a non-exclusive license for you to use
// and redistribute this software, in both source code and binary forms, and the
// accompanying documentation, if any, free of charge, provided that this
// paragraph, and the following two paragraphs appear in all copies and in all
// derivative works.  In no event shall this source code or this license thereto
// impact upon or fall within the scope of any otheragreement that you have with
// Symvionics.
// 
// BY USING THIS SOFTWARE, YOU FURTHER AGREE THAT, IN NO EVENT, SHALL SYMVIONICS
// BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR
// CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
// SOFTWARE, THE ACCOMPANYING DOCUMENTATION, IF ANY, OR ANY DERIVATIVE WORKS.
// YOU AGREE TO DEFEND, INDEMNIFY, AND HOLD HARMLESS SYMVIONICS FROM AND AGAINST
// ALL FINES, PENALTIES, LIABILITIES, DAMAGES, COSTS, AND EXPENSES INCURRED
// BY SYMVIONICS AS A RESULT OF YOUR USE OF THIS SOFTWARE INCLUDING THIRD PARTY
// USE OF ANY DERIVATIVE WORKS CREATED BY YOU.
// 
// SYMVIONICS SPECIFICALLY DISCLAIMS ANY AND ALL WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE.  THE SOFTWARE AND ACCOMPANYING DOCUMENTATION LICENSED
// HEREUNDER ARE PROVIDED "AS IS" WITHOUT WARRANTY OR OBLIGATION OF SUPPORT OR 
// MAINTENANCE.
//
// *****************************************************************************

#ifndef __SIMPLEFUNCTION2_H_
#define __SIMPLEFUNCTION2_H_

#include "resource.h"       // main symbols
#include "IadsFunction.h"

// Example shared variable
static int gSharedNumberBetweenAllInstancesOfThisFunction = 10;


/////////////////////////////////////////////////////////////////////////////
// CSimpleFunction2
class ATL_NO_VTABLE CSimpleFunction2 : 
	public CComObjectRootEx<CComSingleThreadModel>,
	public CComCoClass<CSimpleFunction2, &CLSID_SimpleFunction2>,
  public ISimpleFunction2,
  public IIadsFunction,   // This is the interface used to talk with Iads
  public IErrorInfo       // New in version 6.5. Ability to return error string from custom function
{
public:
	CSimpleFunction2()
	{
	}

DECLARE_REGISTRY_RESOURCEID(IDR_SIMPLEFUNCTION2)

DECLARE_PROTECT_FINAL_CONSTRUCT()

BEGIN_COM_MAP(CSimpleFunction2)
	COM_INTERFACE_ENTRY(ISimpleFunction2)
	COM_INTERFACE_ENTRY(IIadsFunction)
END_COM_MAP()

// ISimpleFunction2
public:

  STDMETHOD(FinalConstruct)( void )
  {
    // The ATL "goo" will call this upon construction of your class. It be called once per class creation.
    // Every derived parameter that get's called and uses this function will create it's own "instance" of this 
    // class, so if you have 10 derived parameters calling the function, this function will be called 10 times, but 
    // each call will be a complete unique copy of this class. 
    // If you want to create any "global" resources, that are shared between all the class instances, make sure you 
    // create a global static variable (i.e. above this class definition see "example shared variable"). Anything that you
    // want kept separate per instance and not shared, declare in the member variable selection below (see CComBSTR mStringOutput below)

    // This variable is for performance and initialization reasons as you will see below in the Compute function. 
    // Make sure to add this member variable to your class -> bool mWasInitialized
    mWasInitialized = false;

    // Preparation for an example on how to output a "blob" data
    // For now, we just set our SafeArray pointer to NULL. We'll do the allocation in the Compute function init section
    mSA = NULL;

    mBlobSizeInBytes = 0;

    return S_OK;
  }

  STDMETHOD(FinalRelease)( void )
  {
    // Do any final memory cleanup here if necessary. 
    // The ATL "goo" will call this upon descruction of your function.
    if ( mSA ) ::SafeArrayDestroy( mSA );

    return S_OK;
  }

  STDMETHOD(Compute)( /*[in]*/ VARIANT* dataIn, /*[out]*/ VARIANT* dataOut )
  {
    // About this function
    // FunctionName in Iads -> SampleFunctionVC.SimpleFunction2  (Basically, this is the "ProgId" for this class)
    // How to call in Iads  -> SampleFunctionVB.SimpleFunction2( "Test", Parameter2, Parameter3 )   (You can also feed in constant values) 
    // In Args              -> StringValue1, FloatValue2, FloatValue3
    // Out Args             -> <Depends on choice below>

    // Background info on the Compute function interface
    // The VARIANT "dataIn" argument contains a "SafeArray" of input arguments passed from the user's
    // function (from of their derived equation) into this function. It will contain all the values
    // that you'll need to compute the function result. The result is returned in the "dataOut" argument.
    // The "dataOut" arguement is a VARIANT so we can return numeric results, as well as strings and other types
    // For a more advanced example on returning strings and other types, see the SimpleFunction2 class in this project
    // This function will be called once per data point calculated

    // Get the input arguments array
    register VARIANT* argsArray = (VARIANT*)(dataIn->parray->pvData);  // Could use SafeArrayAccessData, but slower..

    // The very first call to this function is designated as an "initializion call". 
    // Check all parameter types here and then never do again. Adding this check to your code will speed up the performance 
    if ( !mWasInitialized )
    {
      // The benefit of initializing here is that you can return an error code back to the user in the event of failure
      // Here's a list of possible return values at this point:
      //  A) E_FAIL or E_UNEXPECTED -> Returns an "unspecified error" to the user 
      //  B) E_OUTOFMEMORY -> Returns an "out of memory error" to the user 
      //  C) E_INVALIDARG or DISP_E_TYPEMISMATCH -> If one or more of the function arguments were of incorrect type (i.e. numerical vs. string), returns a "type mismatch" error to the user. 
      //  D) DISP_E_BADPARAMCOUNT -> If the number of function arguments were incorrect, returns an "invalid number of parameters" error to the user 

      // First Step: We need to check how many arguments were passed into our function from the user
      // Example, if the user creates a derived parameter with the function -> SampleFunction.SimpleFunction2( "Format", ValueOrParam1, ValueOrParam2 )
      // The argument count would be 3
      int argCount = dataIn->parray->rgsabound->cElements;
      if ( argCount != 3 )  return DISP_E_BADPARAMCOUNT;

      // Second Step: Check Types of each arg..... Either VT_R8 (floating point value), VT_BSTR (string value) for now...
      // In IADS, most every type of numerical argument is passed via an 8 byte floating point value VT_R8. 
      // If you're in doubt, use VT_R8. You can also break here in the code and examine the argsArray[N].vt value to see the actual type 
      // Example, if the user creates a derived parameter with the function -> SampleFunction.SimpleFunction2( "Format", ValueOrParam1, ValueOrParam2 )
      // The argument type for arg1 would be VT_BSTR... and arg2/arg3 would be VT_R8    
      if ( argsArray[0].vt != VT_BSTR ) return E_INVALIDARG;
      if ( argsArray[1].vt != VT_R8 ) return E_INVALIDARG;
      if ( argsArray[2].vt != VT_R8 ) return E_INVALIDARG;

      // Do whatever kind of alternative initialization you need to here as well. 
      // Examples: Connecting to a TCP socket, Serial Port, or any other type of external device
      //           Connecting to an external database or any external file
      //           Preparing computational lookup tables or anything else to prepare for calculation

      // At this point in time, all Blobs are "fixed size", so you'll need to determine a constant size in bytes and not change it. 
      // In this example, we will create a single blob output array that will hold 2 float values. 
      // The first 4 byte entity is an unsigned integer which will hold the size of the blob (in bytes)
      // The remaining bytes will hold our 5 floating point numbers 
      // The length in bytes would be -> sizeof( unsigned __int32 ) + 5 * sizeof( float )
      const int cNumFloatsInBlob = 5;
      mBlobSizeInBytes = sizeof( unsigned __int32 )/*HeaderSizeInBytes*/ + cNumFloatsInBlob * sizeof( float ) /*Payload*/;

      // Now, let's allocate the SafeArray to contain our blob data 
      mSA = ::SafeArrayCreateVector( VT_UI1, 0, mBlobSizeInBytes ); 
      if ( mSA == NULL ) 
      {
        // Example of returning a custom error string to Iads. See GetDescription below for further info
        mErrorString = "SimpleFunction2 failed to allocate memory for Blob output";
        return E_OUTOFMEMORY;
      }

      mWasInitialized = true;
    }
  


    // Third step: Get the actual values of each arg by extracting from the array of input arguments
    register BSTR arg1 = argsArray[0].bstrVal;
    register double arg2 = argsArray[1].dblVal;
    register double arg3 = argsArray[2].dblVal;

    // Final step: Perform your function's purpose and return the output value
  
    // In this slightly more advanced example, we'll show how to return numeric, string, blobs, and even multi
    enum ReturnExampleType { ReturnDouble=0, ReturnString, ReturnBlob };
  
    // Change this variable to see the different return type examples
    // Also, make sure you adjust your ParameterDefaults "ParamType" column for the derived parameter to match the appropriate output type
    // String = Ascii, Numeric = float/int/uint/long/ulong, Blob = Blob

    ReturnExampleType iWantToSeeAnExampleOfReturning = ReturnBlob;

    if ( iWantToSeeAnExampleOfReturning == ReturnDouble ) 
    {
      // In this example, simply pass back whatever value was given as an argument in arg2. 
      // You could create any value on the input an pass it back at this point.. For instance: A string representation of arg2
      dataOut->vt = VT_R8;
      dataOut->dblVal = arg2 + arg3;
    }
    else if ( iWantToSeeAnExampleOfReturning == ReturnString )
    {
      // In this example, simply pass back whatever string was given as an argument in arg1. 
      // You could create any string based on the input an pass it back at this point.. For instance: A string representation of arg2
      // Hint: Don't forget to set your derived parameter output type (ParamType column in the ParameterDefault table) to "ascii"
      // Hint: Note that the return string is *not* "reallocated" each time per output. All we are returning is a "reference" to our
      //       internally allocated member variable string. This was done for performance consideration purposes
      //       If you do reallocate a return a new string each function call, it will cause memory to leak
      dataOut->vt = VT_BSTR;
      mStringOutput = arg1; 
      dataOut->bstrVal = mStringOutput;
    }
    else // if ( returnType == ReturnBlob )
    {
      // In this example, we'll return the value of arg2 and arg3 in a blob consisting of two 32bit float values
      // A Blob is defined as an array of bytes or VT_ARRAY|VT_UI1 with the first 4 bytes being the size of the array
      // Hint: Don't forget to set your derived parameter output type (ParamType column in the ParameterDefault table) to "blob"
      // Hint: Until Iads version 7.0 is released, you will need to explicitly set your blob output size (size of output array)
      //       within the derived equation using the SetDataOutputSize( NumBytes ) 
      //       Example: SetDataOutputSize(40), SampleFunctionVC.SimpleFunction2( 40, ... )
      // Hint: Note that the return string is *not* "reallocated" each time per output. All we are returning is a "reference" to our
      //       internally allocated member variable string. This was done for performance consideration purposes
      //       If you do reallocate a return a new string each function call, it will cause memory to leak

      // Get a pointer to the safeArray data that we allocated in the "initialization stage" above
      BYTE* sa;
      ::SafeArrayAccessData( mSA, (void**)&sa );

      // Now access the first 4 byte integer so we can inject the Blob size (in bytes). This step will become vital when variable size blobs are implemented
      unsigned __int32* blobSizeInBytes = (unsigned __int32*)sa;

      // Set the blob size in bytes
      *blobSizeInBytes = mBlobSizeInBytes;

      // Now let's inject the data into the remaining part of the Blob. 
      // Get a pointer to the payload portion of the Blob (starting at 5th byte)
      float* payloadValues = (float*)(sa + sizeof(unsigned __int32));

      // Set arg2 and arg3 into the blob payload (array), plus their sum, product, and ratio
      payloadValues[0] = (float)arg2;  
      payloadValues[1] = (float)arg3;
      payloadValues[2] = (float)(arg2+arg3);  
      payloadValues[3] = (float)(arg2*arg3);
      payloadValues[4] = (float)(arg2/arg3);  

      // Likewise, if you had your own internal function to compute the results, you could simply pass in the input args
      // and a reference to this array. Your function would simply write the result directly into the array
      // Make sure you maintain a consistent order to your outputs, because we'll have to extract them "one by one" later
      // CalculateMyResults( arg2, arg3, payloadValue );

      SafeArrayUnaccessData( mSA );

      // Finally, set the blob output type and reference the safeArray we just built 
      dataOut->vt = VT_ARRAY|VT_UI1;
      dataOut->parray = mSA;
    }

    return S_OK;
  }

  //////////////////////////////
  // Optional IErrorInfo interface implementation. If an error code is returned above from 
  // the !mWasInitialized code section above, and this interface is implemented, Iads will 
  // call this method below to obtain a string decription of the error encounterd (and present it to the user)
  STDMETHOD(GetDescription)( BSTR* pVal )
  {
    if ( !pVal ) return E_POINTER;
    *pVal = mErrorString.Copy();
    return S_OK;
  }
  STDMETHOD(GetGUID)( GUID* pVal ){ return E_NOTIMPL; }
  STDMETHOD(GetHelpContext)( DWORD* pVal ) { return E_NOTIMPL; }
  STDMETHOD(GetHelpFile)( BSTR* pVal ) { return E_NOTIMPL; }
  STDMETHOD(GetSource)( BSTR* pVal ) { return E_NOTIMPL; }
  //////////////////////////////

  // Member variables of this custom function
  CComBSTR mStringOutput;
  CComBSTR mErrorString;
  SAFEARRAY* mSA;
  SAFEARRAY* mSADouble;
  bool mWasInitialized;
  int mBlobSizeInBytes;

};

#endif //__SIMPLEFUNCTION2_H_
